"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var isClass = require("is-class");
/**
 * Utility function that returns the full chain of prototypes (excluding Object.prototype) from the given prototype.
 * The order will be [proto, protoParent, protoGrandparent, ...]
 */
function getProtoChain(proto) {
    var protoChain = [];
    while (proto !== Object.prototype) {
        protoChain.push(proto);
        proto = Object.getPrototypeOf(proto);
    }
    return protoChain;
}
/**
 * Utility function that works like `Object.apply`, but copies properties with getters and setters properly as well.
 * Additionally gives the option to exclude properties by name.
 */
function copyProps(dest, src, exclude) {
    if (exclude === void 0) { exclude = []; }
    var props = Object.getOwnPropertyDescriptors(src);
    for (var _i = 0, exclude_1 = exclude; _i < exclude_1.length; _i++) {
        var prop = exclude_1[_i];
        delete props[prop];
    }
    Object.defineProperties(dest, props);
}
/**
 * Utility function for mixing a number of prototypes into a target object.  Utilizes getProtoChain() to intelligently
 * apply the mixing.
 */
function mixPrototypes(target, ingredients, exclude) {
    if (exclude === void 0) { exclude = []; }
    var appliedPrototypes = [];
    for (var _i = 0, ingredients_1 = ingredients; _i < ingredients_1.length; _i++) {
        var prototype = ingredients_1[_i];
        var protoChain = getProtoChain(prototype);
        // Apply the prototype chain in reverse order, so that old methods don't override newer ones; also make sure
        // that the same prototype is never applied more than once.
        for (var i = protoChain.length - 1; i >= 0; i--) {
            var newProto = protoChain[i];
            if (appliedPrototypes.indexOf(newProto) === -1) {
                copyProps(target, protoChain[i], exclude);
                appliedPrototypes.push(newProto);
            }
        }
    }
}
function Mixin() {
    var ingredients = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        ingredients[_i] = arguments[_i];
    }
    // Start building a class that represents the mixture of the given Base and Class
    var Mixed = /** @class */ (function () {
        function Mixed() {
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            for (var _a = 0, ingredients_2 = ingredients; _a < ingredients_2.length; _a++) {
                var constructor = ingredients_2[_a];
                // If the constructor is a callable JS function, we would prefer to apply it directly to `this`,
                if (!isClass(constructor))
                    constructor.apply(this, args);
                // but if it's an ES6 class, we can't call it directly so we have to instantiate it and copy props
                else
                    copyProps(this, new (constructor.bind.apply(constructor, [void 0].concat(args)))());
            }
        }
        return Mixed;
    }());
    // Apply prototypes, including those up the chain
    mixPrototypes(Mixed.prototype, ingredients.map(function (ctor) { return ctor.prototype; }), ['constructor']);
    // Mix static properties
    mixPrototypes(Mixed, ingredients, ['prototype', 'length', 'name']);
    return Mixed;
}
exports.Mixin = Mixin;
/**
 * A decorator version of the `Mixin` function.  You'll want to use this instead of `Mixin` for mixing generic classes.
 */
var mix = function () {
    var ingredients = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        ingredients[_i] = arguments[_i];
    }
    // @ts-ignore
    return function (decoratedClass) { return Mixin.apply(void 0, (ingredients.concat([decoratedClass]))); };
};
exports.mix = mix;
